#include "GTXLReader.h"
#include <iostream>

#include <cstdio>
// #include <cstring>
#include <cstdlib>

#include <stdexcept>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

  GTXLReader::GTXLReader(char* filename)
  {
    doc = xmlParseFile(filename);
	
    if (doc == NULL ) 
      {
	fprintf(stderr,"Document not parsed successfully. \n");
	xmlFreeDoc(doc);
	throw runtime_error("XML parser error");
      }
	
	
    if (xmlStrcmp(xmlDocGetRootElement(doc)->name, (const xmlChar *) "GTS")) 
      {
	fprintf(stderr,"document of the wrong type, root node != GTS");
	xmlFreeDoc(doc);
	throw runtime_error("Wrong document type");
      }

  } 
  GTXLReader::~GTXLReader()
  {
    xmlFreeDoc(doc);
  }

  GraphGrammar GTXLReader::read()
  {
    xmlNodePtr root = xmlDocGetRootElement(doc);
    xmlNodePtr cur = root->xmlChildrenNode;
    
    //variables to check id's unique
    set<string> id_checker1;
    pair<set<string>::iterator, bool> checker1;
    string id_string;
    
    
    while(NULL != cur)
      {
	if(!xmlStrcmp(cur->name,(const xmlChar *)"Initial")) // initial graph?
	  {
	    if(0) // is there already an initial graph?
	      {
		cerr << "There is more than one initial graph specified in the input file.\n";
		throw runtime_error("More than one initial graph specified");
	      }
	    else // first initial graph?
	      {
		// Initial contains a Graph node
		xmlNodePtr igraph = cur->xmlChildrenNode;
		while(NULL!=igraph && xmlStrcmp(igraph->name,(const xmlChar *)"Graph"))
		  igraph = igraph->next;
		if(NULL==igraph)
		  {
		    cerr << "Initial without Graph section found.\n";
		    throw runtime_error("Initial without Graph section found.");
		  }
		else
		{  
		//check id's unique
		// printf("name = %s\n",(char*)igraph->name);
	  	id_string = (char*)xmlGetProp(igraph,(const xmlChar *)"id");
	  	checker1 = id_checker1.insert(id_string);
		if (checker1.second == false)
		{
			printf("Error: The id's in 'Graph' are not unique: id = %s\n", id_string.c_str());
			exit(0);
		} 
		}
		
		map<string,IDTYPE> dummy;
		initialGraph = readHypergraph(igraph,dummy);
	      }
	    
	  }
	else if(!xmlStrcmp(cur->name,(const xmlChar *)"Rule")) // transformation rule?
	  {
	   	//check id's unique
		// printf("name = %s\n",(char*)cur->name);
	  	id_string = (char*)xmlGetProp(cur,(const xmlChar *)"id");
	  	checker1 = id_checker1.insert(id_string);
		if (checker1.second == false)
		{
			printf("Error: The id's in 'Rule' are not unique: id = %s\n", id_string.c_str());
			exit(0);
		}
	  
	    TransformationRule r = readRule(cur);
	    rules.push_back(r);
	  }
	else
	  {
	    //cerr << "Ignoring XML node " << cur->name << "\n";
	  }
	cur = cur->next;
      }

    if(0) // no initial graph specified
      {
	cerr << "Input file does not specify an initial graph.\n";
	throw runtime_error("No initial graph specified");
      }

    GraphGrammar result(initialGraph);
    
     // cout << rules.size() << endl;
    
    for(list<TransformationRule>::iterator i = rules.begin();
	i!= rules.end();
	i++)
      {
      	TransformationRule r(*i);
	result.addRule(r);
      }
    return result;
  }
  


  Hypergraph GTXLReader::readHypergraph(xmlNodePtr root,std::map<std::string,IDTYPE>& xmlID2ID)
  {
    // true if hypergraph=true is set
    bool isHypergraph(!xmlStrcmp(xmlGetProp(root,(const xmlChar *)"hypergraph"),(const xmlChar *)"true"));
    const xmlChar *edgemodeName(xmlGetProp(root,(const xmlChar *)"edgemode"));
    
    if( isHypergraph 
       && !xmlStrcmp(edgemodeName,(const xmlChar *)"directed"))
      {
	cerr << "This program does not support directed hypergraphs.\n";
	throw runtime_error("directed hypergraph in input file.");
      }

    bool isDirected(false);
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"directed")) 
      isDirected = true;
    if(!xmlStrcmp(edgemodeName,(const xmlChar *)"defaultdirected"))
      isDirected = true;

    Hypergraph result;
    
    //variables to check id's unique
    set<string> id_checker;
    pair<set<string>::iterator, bool> checker;
    string id_string;
    
    // start parsing the node/edge list:
    xmlNodePtr cur = root->xmlChildrenNode;
    while(NULL != cur)
      {
	// Node:
	if(!xmlStrcmp(cur->name,(const xmlChar *)"node"))
	  {
	  
	  //check node's id's
	  	id_string = (char*)xmlGetProp(cur,(const xmlChar *)"id");
	  	checker = id_checker.insert(id_string);
		if (checker.second == false)
		{
			printf("Error: The id's in 'node' are not unique: id = %s\n", id_string.c_str());
			exit(0);
		}
		
	  } // do nothing; unconnected nodes are not permitted in simple transformation rules, and connected nodes are implied by IDs stored in edges
	// (biconnected) edge
	else if(!xmlStrcmp(cur->name,(const xmlChar *)"edge"))
	  {
	  
	  //check edge's id's
	  	id_string = (char*)xmlGetProp(cur,(const xmlChar *)"id");
	  	checker = id_checker.insert(id_string);
		if (checker.second == false)
		{
			printf("Error: The id's in 'edge' are not unique: id = %s\n", id_string.c_str());
			exit(0);
		}
		
		
	    xmlChar* fromXMLID = xmlGetProp(cur,(const xmlChar *)"from");
	    xmlChar* toXMLID = xmlGetProp(cur,(const xmlChar *)"to");

	    IDTYPE fromID,toID;
	    if(1==xmlID2ID.count(string((char*)fromXMLID)))
	      {
		fromID = xmlID2ID[string((char*)fromXMLID)];
		result.addVertex(fromID);
	      }
	    
	    else
	      {
		fromID = createNewID();
		result.addVertex(fromID);
		xmlID2ID[string((char*)fromXMLID)] = fromID;
	      }
	    if(1==xmlID2ID.count(string((char*)toXMLID)))
	      {
		toID = xmlID2ID[string((char*)toXMLID)];
		result.addVertex(toID);
	      }
	    
	    else
	      {
		toID = createNewID();
		result.addVertex(toID);
		xmlID2ID[string((char*)toXMLID)] = toID;
	      }
	    // find out if isdirected attribute is specified. If not, use default. Undirected Edges are represented by two directed edges.
	    bool edgeIsDirected(isDirected);
	    if(!xmlStrcmp(xmlGetProp(cur,(const xmlChar *)"isdirected"),(const xmlChar *)"true")) edgeIsDirected = true;
	    if(!xmlStrcmp(xmlGetProp(cur,(const xmlChar *)"isdirected"),(const xmlChar *)"false")) edgeIsDirected = false;
	    
	    string label(getLabel(cur));
	    
	    
	    if(""==label) // warn if unlabeled edge was found
	      {
		cerr << "WARNING: unlabeled edge found, using empty label.\n";
	      }
	    list<IDTYPE> connectionList;
	    connectionList.push_back(fromID);
	    connectionList.push_back(toID);
	    IDTYPE newEdgeID = result.addEdge(connectionList,label);
	    xmlChar *edgeXMLID = xmlGetProp(cur,(const xmlChar *)"id");
	    if(NULL != edgeXMLID)
	      xmlID2ID[string((char*)edgeXMLID)] = newEdgeID;
	    if(!edgeIsDirected)
	      {
		connectionList.clear();
		connectionList.push_back(toID);
		connectionList.push_back(fromID);
		newEdgeID = result.addEdge(connectionList,label);
		xmlChar *edgeXMLID = xmlGetProp(cur,(const xmlChar *)"id");
		if(NULL != edgeXMLID)
		  xmlID2ID[string((char*)edgeXMLID)] = newEdgeID;

	      }
	  }
	// Hyperedge
	else if(!xmlStrcmp(cur->name,(const xmlChar *)"rel"))
	  {
	  	//check id's in rel
	  	id_string = (char*)xmlGetProp(cur,(const xmlChar *)"id");
	  	checker = id_checker.insert(id_string);
		if (checker.second == false)
		{
			printf("Error: The id's in rel are not unique: id = %s\n", id_string.c_str());
			exit(0);
		}
	  
	    if(!isHypergraph) // only hypergraphs have rel edges
	      {
		cerr << "Hyperedge found in a non-hypergraph\n";
		throw runtime_error("Hyperedge found in a non-hypergraph");
	      }
	    
/*	    if(!xmlStrcmp(xmlGetProp(cur,(const xmlChar *)"isdirected"),(const xmlChar *)"true"))
	      {
		// directed hyperedges are unsupported
		cerr << "Directed Hyperedge found. Directed Hyperedges are not supported.\n";
		throw runtime_error("Directed Hyperedges are unsupported.");
	      }
*/
	    list<IDTYPE> connectionList;
	    xmlNodePtr relend = cur->xmlChildrenNode;
	    while(NULL != relend)
	      {
		if(!xmlStrcmp(relend->name,(const xmlChar *)"relend"))
		  {
		  	// check id's in relend
			id_string = (char*)xmlGetProp(relend,(const xmlChar *)"id");
	  		checker = id_checker.insert(id_string);
			if (checker.second == false)
			{
				printf("Error: The id's in relend are not unique: id = %s\n", id_string.c_str());
				exit(0);
			}
			
		    xmlChar* targetXMLID = xmlGetProp(relend,(const xmlChar *)"target");
		    IDTYPE targetID;
		    if(1==xmlID2ID.count(string((char*)targetXMLID)))
		      {
			targetID = xmlID2ID[string((char*)targetXMLID)];
			result.addVertex(targetID);
		      }
		    
		    else
		      {
			targetID = createNewID();
			result.addVertex(targetID);
			xmlID2ID[string((char*)targetXMLID)] = targetID;
		      }
		    connectionList.push_back(targetID);

		  }
		relend = relend->next;
	      }
	
	    string label = getLabel(cur);
	    
	    
	    if(""==label) // warn if unlabeled edge was found
	      {
		cerr << "WARNING: unlabeled hyperedge found, using empty label.\n";
	      }

	    IDTYPE newEdgeID = result.addEdge(connectionList,label);
	    xmlChar *edgeXMLID = xmlGetProp(cur,(const xmlChar *)"id");
	    if(NULL != edgeXMLID)
	      xmlID2ID[string((char*)edgeXMLID)] = newEdgeID;

	  }
	cur = cur->next;
      }

    // now set to the hypergraph original string IDs
    for(map<string, IDTYPE>::iterator it=xmlID2ID.begin(); it!=xmlID2ID.end(); it++)
       result.setIdLabel(it->second, it->first);
      
    return result;
  }


  TransformationRule GTXLReader::readRule(xmlNodePtr node)
  {
    // first, find the "mapping" child element:
    xmlNodePtr cur = node->xmlChildrenNode;
    while(NULL != cur && xmlStrcmp(cur->name,(const xmlChar *)"Mapping"))
      cur = cur->next;

    if(NULL == cur) // no mapping specified?
      {
	cerr << "Transformation rule without mapping found\n";
	throw runtime_error("Transformation rule without mapping found\n");
      }
    // for each map element, register one ID for later use in hypergraph
    // construction
    map<string,IDTYPE> xmlID2ID;
    for(xmlNodePtr mapEl = cur->xmlChildrenNode;
	NULL != mapEl;
	mapEl = mapEl->next)
      {
	// ignore everything except MapElem nodes
	if(!xmlStrcmp(mapEl->name,(const xmlChar *)"MapElem"))
	  {
	    string fromXmlID((char*)xmlGetProp(mapEl,(const xmlChar *)"from"));
	    string toXmlID ((char*)xmlGetProp(mapEl,(const xmlChar *)"to"));
	    IDTYPE id = createNewID();
	    xmlID2ID[fromXmlID] = id;
	    xmlID2ID[toXmlID] = id;
	  }
      }
    // map reading complete.
    // now read left hand side and right hand side
    Hypergraph lhs;
    Hypergraph rhs;
    for(cur = node->xmlChildrenNode; NULL != cur; cur = cur->next)
      {
	if(!xmlStrcmp(cur->name,(const xmlChar *)"LHS"))
	  lhs = readRuleGraph(cur,xmlID2ID);
	if(!xmlStrcmp(cur->name,(const xmlChar *)"RHS"))
	  rhs = readRuleGraph(cur,xmlID2ID);
      }

    // get xml-ID of the rule:
    string strRuleID;
    xmlChar *ruleid = xmlGetProp(node,(const xmlChar *)"id");
    if(NULL != ruleid)
      {
	strRuleID = string((char *)ruleid);
      }
      
    return TransformationRule(lhs,rhs,split_space(strRuleID));
  }

  Hypergraph GTXLReader::readRuleGraph(xmlNodePtr node, std::map<std::string,IDTYPE>& xmlID2ID)
  {
    // a rule graph contains a (hyper-)graph and some information which is
    // ignored here.
    xmlNodePtr cur = node->xmlChildrenNode;
    while(NULL != cur && xmlStrcmp(cur->name,(const xmlChar *)"RuleGraph"))
      cur = cur->next;
    if(NULL==cur) // no RuleGraph in the LHS/RHS?
      {
	cerr << node->name << " element without RuleGraph element found.\n";
	throw runtime_error("LHS/RHS element without RuleGraph element found.\n");
      }

    cur = cur->xmlChildrenNode;
    
    while(NULL != cur && xmlStrcmp(cur->name,(const xmlChar *)"Graph"))
      cur = cur->next;

    if(NULL==cur) // no graph in the RuleGraph?
      {
	cerr << "RuleGraph element without a graph found.\n";
	throw runtime_error("RuleGraph element without a graph found.\n");
      }

    return readHypergraph(cur, xmlID2ID);
  }

  std::string GTXLReader::getLabel(xmlNodePtr node)
  {
    for(xmlNodePtr cur = node->xmlChildrenNode; NULL!=cur; cur = cur->next)
      {
	
	if(!xmlStrcmp(cur->name,(const xmlChar *)"attr"))
	  if(!xmlStrcmp(xmlGetProp(cur,(const xmlChar *)"name"),(const xmlChar *)"label"))
	    {
	      cur = cur->xmlChildrenNode;
	      while(NULL!=cur && xmlStrcmp(cur->name,(const xmlChar *)"string"))
		cur = cur->next;
	      if(NULL==cur) return "";
	      else return string((char*)cur->xmlChildrenNode->content);
	    }
      }
    
    return "";
    
  }
 
string GTXLReader::split_space(string s)
{
    char * pch;
    string res = "";
    char * ch=new char(s.size());
    strcpy(ch, s.c_str());
    pch=strtok(ch, " ");
    res += (string)pch;
    while(pch!=NULL)
    {
    	pch=strtok(NULL, " ");
	if(pch!=NULL)
	 res += (string)pch;
    }
    
    return res;
}
