
#include "GXLReader.h"
#include "GTXLReader.h"

#include "Globals.h"

#include <stdexcept>

GXLReader::GXLReader(char * approxfilename)
{
  doc = xmlParseFile(approxfilename);

  maxID = 0;

  if (doc == NULL )
	{
		fprintf(stderr,"Document not parsed successfully. \n");
		xmlFreeDoc(doc);
	}
}	

GXLReader::~GXLReader()
{
}

/**
 * returns true if the given node has an user attribute (ATTR) with the specified name,
 * false otherwise
 */
bool GXLReader::hasUserAttr(xmlNodePtr cur, const char* name)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }
  return (cur != NULL);
}

/**
 * returns the value of the user attribute (ATTR) of the node with the specified name of the
 * specified type or NULL if noch such attribute exists
 */
const char* GXLReader::getUserAttrValue(xmlNodePtr cur, const char* name, const char* type)
{
  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if ( !xmlStrcasecmp(cur->name, (const xmlChar*) "attr") &&
         !xmlStrcasecmp((const xmlChar*) name, xmlGetProp(cur, (const xmlChar*) "name")) )
    {
      break;
    }
    else
    {
      cur = cur->next;
    }
  }

  if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
  {
    cur = cur->xmlChildrenNode;
    while ( (cur != NULL) && xmlStrcasecmp((const xmlChar*) type, cur->name) )
    {
      cur = cur->next;
    }
    if ( (cur != NULL) && (cur->xmlChildrenNode != NULL) )
    {
      return (const char*) cur->xmlChildrenNode->content;
    }
  }
  return NULL;
}

/**
 * processes the node cur is pointing
 */
void GXLReader::processNode(xmlNodePtr cur)
{
  IDTYPE lastUsedID = getLastID();

  try
  {
  const char* id_ = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  if (id_ == NULL)
     throw runtime_error("Node id expected");

  char * id = strtok((char*)id_, "_");

  if(atoi(id)>maxID)
			maxID = atoi(id);

  Hypergraph hg_start = ts.getStartGraph();
     
  // vertex   
  if(hasUserAttr(cur, "vertex"))
  {
		hg.addVertex(atoi(id)+lastUsedID);

		// iota mapping
    const char* iota_string = getUserAttrValue(cur, "iota_mapping", "string");
    if (iota_string != NULL)
    {
			char * ch;
			ch = strtok((char*)iota_string, ";");

			while(ch!=NULL)
			{
				bool found;
				IDTYPE id_start = hg_start.getLabelId(ch, &found);
				if(found)
					iota[id_start] = atoi(id)+lastUsedID;
				else
					throw runtime_error("spoiled original label");
				ch = strtok(NULL, ";");
			}
		}
 
		return;
	}
  
  // transition
  if(hasUserAttr(cur, "transition"))
  {
     const char* rule = getUserAttrValue(cur, "rule", "string");
     if (rule == NULL)
			 throw runtime_error("Transition without rule");

		 IDTYPE trans_id = atoi(id)+lastUsedID;
		 list<PTNetTransition>::iterator it;
		 for(it = transitions.begin(); it!=transitions.end(); it++)
		 {
			 if(it->getID()==trans_id)
				 break;
		 }

		 string rname = rule;
		 		 
     bool found;
		 int id_rule = ts.findRule(split_space(rname).c_str(), &found);
		 if(!found)
			 	 throw runtime_error("name of the rule doesn't exist");
		 
		 // new transition
		 if(it==transitions.end())
		 {
			 PTNetTransition pnt;
       pnt.setID(trans_id);
       pnt.setAssociatedRule(id_rule);
			 transitions.push_back(pnt);
		 }
		 else
				it->setAssociatedRule(id_rule);

		Hypergraph hleft = ts.getRule(id_rule).getLeft();
		Hypergraph hright = ts.getRule(id_rule).getRight();

		map<IDTYPE, IDTYPE> mu_map;

		 // mu mapping
    const char* mu_string = getUserAttrValue(cur, "mu_mapping", "string");
    if (mu_string == NULL)
			 throw runtime_error("transitions without mu mapping are not allowed \n Probably you use old version of unfolding program");

		char * ch1;
		ch1 = strtok((char*)mu_string, ",;");
    char * ch2;
    ch2 = strtok(NULL, ",;");

		while((ch1!=NULL) && (ch2!=NULL))
		{
      bool found;
			IDTYPE id_ = hleft.getLabelId(ch1, &found);
			if(!found)
				id_ = hright.getLabelId(ch1, &found);
			else if(!found)
					throw runtime_error("spoiled original label");
			
      if(atoi(ch2)>maxID)
				maxID = atoi(ch2);
				
      mu_map[id_] = atoi(ch2)+lastUsedID;

			ch1 = strtok(NULL, ",;");
			ch2 = strtok(NULL, ",;");
		}

		mu[trans_id]=mu_map;
	}
	}
	catch(exception e)
	{
		cout << "error by reading the node" << endl;
		throw e;
	}

}

/**
 * processes the rel cur is pointing
 */
void GXLReader::processRel(xmlNodePtr cur)
{
	IDTYPE lastUsedID = getLastID();

	try
	{
  const char* id_ = (const char*) xmlGetProp(cur, (const xmlChar*) "id");
  
  if (id_ == NULL)
     throw runtime_error("Rel id expected");

  char * id = strtok((char*)id_, "_");

	if(atoi(id)>maxID)
			maxID = atoi(id);

  const char* label = getUserAttrValue(cur, "label", "string");
  if (label == NULL)
  {
    label = "";
  }

  const char* in_marking = getUserAttrValue(cur, "initial_marking", "int");
  if (in_marking == NULL)
		throw runtime_error("initial marking expected");

  initial_marking[atoi(id)+lastUsedID] = atoi(in_marking);
	
  //remove blanks and store the label of the place
  string s = label;
  while (s.find(" ") != string::npos)
    s.replace(s.find(" "), 1, "");

  Hypergraph hg_start = ts.getStartGraph();  

  const char* iota_mapping = getUserAttrValue(cur, "iota_mapping", "string");

  if (iota_mapping != NULL)
  {
		char * ch;
		ch = strtok((char*)iota_mapping, ";");
      
      while(ch!=NULL)       
      {
				bool found;
				IDTYPE id_start = hg_start.getLabelId(ch, &found);
				if(found)
					iota[id_start] = atoi(id)+lastUsedID;
				else
				{
					throw runtime_error("spoiled original label");
				}
				ch = strtok(NULL, ";");
			}
	}	

  // nodes  
  list<IDTYPE> nodes;

  cur = cur->xmlChildrenNode;
  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "relend"))
    {
      const xmlChar* role = xmlGetProp(cur, (const xmlChar*) "role");
      // preset, postset
      if ( (!xmlStrcasecmp(role, (const xmlChar*) "preset")) ||
           (!xmlStrcasecmp(role, (const xmlChar*) "postset")) )
      {
        const char* target = (const char*) xmlGetProp(cur, (const xmlChar*) "target");

        const char* weight = getUserAttrValue(cur, "weight", "int");
        if (target == NULL)
        {
          throw runtime_error("Relend target expected");
        }
        if (weight == NULL)
        {
          weight = "1";
        }
	
        target = strtok((char*)target, "_");  
        int w = atoi(weight);

        if(atoi(target)>maxID)
					maxID = atoi(target);

        IDTYPE trans_id = atoi(target)+lastUsedID;
        list<PTNetTransition>::iterator it;
        for(it = transitions.begin(); it!=transitions.end(); it++)
        {
					if(it->getID()==trans_id)
					break;
				}

				if (!xmlStrcasecmp(role, (const xmlChar*) "preset"))
        {
          // new transition
          if(it==transitions.end())
          {
						PTNetTransition pnt;
						pnt.setID(trans_id);
						pnt.addToPreset(atoi(id)+lastUsedID, w);
						transitions.push_back(pnt);
					}
					else
						it->addToPreset(atoi(id)+lastUsedID, w);
        }
        else
        {
          // new transition
          if(it==transitions.end())
          {
						PTNetTransition pnt;
						pnt.setID(trans_id);
						pnt.addToPostset(atoi(id)+lastUsedID, w);
						transitions.push_back(pnt);
					}
					else
						it->addToPostset(atoi(id)+lastUsedID, w);
				}
      }
      
      // vertex
      if(!xmlStrcasecmp(role, (const xmlChar*) "vertex"))
      {
					const xmlChar* target_ = xmlGetProp(cur, (const xmlChar*) "target");
//				const xmlChar* startorder = xmlGetProp(cur, (const xmlChar*) "startorder");
					
					char *target = strtok((char*)target_, "_");

					if(atoi(target)>maxID)
						maxID = atoi(target);
					
					nodes.push_back(atoi(target)+lastUsedID);
      }
    }
    cur = cur->next;
  }

  // set edge to the hypergraph
  hg.addEdge(nodes, s, atoi(id)+lastUsedID);

  }
  
  catch(exception e)
  {
		 cout << "error by reading the relation" << endl;
		 throw e;
  }
}

Petrigraph GXLReader::read(long * addID)
{
	try
	{
  xmlNodePtr cur;

  cur = xmlDocGetRootElement(doc);
  if (cur == NULL)
      throw runtime_error("Empty xml document");

  if (xmlStrcasecmp(cur->name, (const xmlChar*) "gxl"))
      throw runtime_error("Root node 'GXL' expected");

  cur = cur->xmlChildrenNode;
  if (cur == NULL)
        throw runtime_error("Graph expected");

  while ( (cur != NULL) && xmlStrcasecmp(cur->name, (const xmlChar*) "graph") )
  {
     cur = cur->next;
  }

  if (cur == NULL)
  {
     throw runtime_error("Graph expected");
  }      

  cur = cur->xmlChildrenNode;

  while (cur != NULL)
  {
    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "node"))
    		processNode(cur);

    if (!xmlStrcasecmp(cur->name, (const xmlChar*) "rel"))
				processRel(cur);

		cur = cur->next;
	}        
      
  PTNet ptn(hg.getEdges(), transitions, initial_marking);    
  Petrigraph pg(hg, ptn, ts, mu, iota);
  pg.setIota(iota);

  (*addID) = getLastID();

  setLastID(getLastID()+maxID);
    
  return pg;
  
  }
  
  catch(exception e)
  {
		cout << "error by reading GXL file" << endl;
		throw e;
  }
}

/** set graph grammar */
void GXLReader::setGraphGrammar(GraphGrammar gr)
{
	ts = gr;
}


/*!
    \fn GXLReader::split_space(string s)
 */
string GXLReader::split_space(string s)
{
    char * pch;
    string res = "";
    char ch[256];
    strcpy(ch, s.c_str());
    pch=strtok(ch, " ");
    res += (string)pch;
    while(pch!=NULL)
    {
    	pch=strtok(NULL, " ");
	if(pch!=NULL)
	 res += (string)pch;
    }
    
    return res;
}
