#include "PTNet.h"
#include "Globals.h"
#include <iostream>
#include <stdexcept>
#include <stack>

using namespace std;

  PTNetTransition::PTNetTransition(std::set<IDTYPE> p_preset,
                                   std::set<IDTYPE> p_postset,
                                   std::map<IDTYPE,unsigned> p_preweight,
                                   std::map<IDTYPE,unsigned> p_postweight,
                                   IDTYPE p_associatedRule)
    : preset(p_preset),
      postset(p_postset),
      preweight(p_preweight),
      postweight(p_postweight),
      associatedRule(p_associatedRule)
  {
    id = createNewID();
  }
  
  PTNetTransition::PTNetTransition(const PTNetTransition& t)
  {
    preset = t.getPreset();
    postset = t.getPostset();
    preweight = t.preweight;
    postweight = t.postweight;
    associatedRule = t.getAssociatedRule();
    id = t.getID();
  }

  PTNetTransition::PTNetTransition()
  {

		
  }
  

  bool PTNetTransition::isEnabledBy(PTMarking marking) const
  {
    for(set<IDTYPE>::iterator i = preset.begin();
        i!=preset.end();
        i++)
    {
      unsigned arc_weight = getPreWeight(*i);
      unsigned n_tokens = (*(marking.find(*i))).second;
      if(arc_weight > n_tokens)
        return false;
    }
    return true;
  }

  PTMarking PTNetTransition::fire(const PTMarking& marking) const
  {
    assert(isEnabledBy(marking));
    PTMarking result;
    for(PTMarking::const_iterator i = marking.begin();
        i!=marking.end();
        i++)
    {
      // number of tokens on the place before fireing
      unsigned tokens = (*i).second; // precondition ensures tokens>=0
      // ...plus added tokens in the postset of *this
      if(1==postset.count((*i).first))
      {
        if(OMEGA == tokens) // handle "infinity" case
          tokens = OMEGA;
        else
          tokens += getPostWeight((*i).first);
      }

      // ... minus removed tokens in the preset of *this
      if(1==preset.count((*i).first))
      {
        if(OMEGA==tokens)
          tokens = OMEGA;
        else
          tokens -= getPreWeight((*i).first);
      }
      result.insert(pair<IDTYPE,unsigned>((*i).first,tokens));
    }
    return result;
  }
  

  
  std::set<IDTYPE> PTNetTransition::getPreset() const
  {
    return preset;
  }
  
  std::set<IDTYPE> PTNetTransition::getPostset() const
  {
    return postset;
  }
  
  unsigned PTNetTransition::getPreWeight(IDTYPE place) const
  {
    if(1==preweight.count(place)) 
      return (*(preweight.find(place))).second;
    else
      return 0;
  }
  
  unsigned PTNetTransition::getPostWeight(IDTYPE place) const
  {
    if(1==postweight.count(place))
      return (*(postweight.find(place))).second;
    else
      return 0;
  }

  IDTYPE PTNetTransition::getID() const
  {
    return id;
  }
  
  IDTYPE PTNetTransition::getAssociatedRule() const
  {
    return associatedRule;
  }
  

  ostream& operator<<(ostream& os, PTNetTransition& trans)
  {
    os << "transition " << trans.id << ": pre{";
    for(map<IDTYPE,unsigned>::iterator i = trans.preweight.begin();
        i!=trans.preweight.end();
        i++)
      os << (*i).first << "(" << (*i).second << ") ";

    os << "}  post{";

    for(map<IDTYPE,unsigned>::iterator i = trans.postweight.begin();
        i!=trans.postweight.end();
        i++)
      os << (*i).first << "(" << (*i).second << ") ";

    os << "}\n";
    return os;
  }
  
  void CovGraph::clear()
  {
    m_nodeMapping.clear();
    m_markingMapping.clear();
    m_succ.clear();
    m_pred.clear();
    m_work.clear();
  }
  
  void CovGraph::addNode(const PTMarking& marking)
  {
    if (m_nodeMapping.find(marking) != m_nodeMapping.end())
      throw invalid_argument("CovGraph::addNode() - Node already inserted");

    IDTYPE newID = m_nodeMapping.size();
    m_nodeMapping[marking] = newID;
    m_markingMapping[newID] = marking;
  }
  
  void CovGraph::addEdge(const PTMarking& m1, const PTMarking& m2, const PTNetTransition& transition)
  {
    map<PTMarking,IDTYPE>::iterator it1 = m_nodeMapping.find(m1);
    map<PTMarking,IDTYPE>::iterator it2 = m_nodeMapping.find(m2);
    
    if ( (it1 == m_nodeMapping.end()) || (it2 == m_nodeMapping.end()) )
      throw invalid_argument("CovGraph::addEdge() - No such node");
      
    IDTYPE id1 = it1->second;
    IDTYPE id2 = it2->second;
    
    m_succ[id1][id2] = transition.getID();
    m_pred[id2][id1] = transition.getID();
  }
  
  bool CovGraph::containsNode(const PTMarking& marking)
  {
    return (m_nodeMapping.find(marking) != m_nodeMapping.end());
  }
  
  bool CovGraph::getNodeWithSmallerMarking(const PTMarking& node, const PTMarking& marking, PTMarking& smallerNode)
  {
    if (m_nodeMapping.find(node) == m_nodeMapping.end())
      throw invalid_argument("CovGraph::containsNodeWithSmallerMarking() - No such node");
      
    map<IDTYPE,bool> visited;
    for(map<IDTYPE,PTMarking>::iterator i = m_markingMapping.begin(); i != m_markingMapping.end(); i++)
      visited[i->first] = false;
      
    IDTYPE actualNode = m_nodeMapping[node];
    stack<IDTYPE> work;
    
    bool finished = false;
    
    while (! finished)
    {
      if (isStrictlyLower(m_markingMapping[actualNode], marking))
      {
        smallerNode = m_markingMapping[actualNode];
        return true;
      }
      for(map<IDTYPE,IDTYPE>::iterator i = m_pred[actualNode].begin(); i != m_pred[actualNode].end(); i++)
      {
        if (! visited[i->first])
        {
          work.push(i->first);
          visited[i->first] = true;
        }
      }
      if (! work.empty())
      {
        actualNode = work.top();
        work.pop();
      }
      else
        finished = true;       
    }
    return false;
  }

  ostream& operator<<(ostream& os, CovGraph& cg)
  {
    os << "Nodes:\n";
    for (map<PTMarking,IDTYPE>::iterator i = cg.m_nodeMapping.begin(); i != cg.m_nodeMapping.end(); i++)
    {
      PTMarking marking = i->first;
      os << "Node " << i->second << ": " << marking <<"\n";
    }
    os << "\nEdges:\n";
    for (map<IDTYPE,map<IDTYPE,IDTYPE> >::iterator it1 = cg.m_succ.begin(); it1 != cg.m_succ.end(); it1++)
    {
      for (map<IDTYPE,IDTYPE>::iterator it2 = it1->second.begin(); it2 != it1->second.end(); it2++)
      {
        os << it1->first << " ==> " << it2->first << "\n";
      }
    }
    return os;
  }
  
  list<PTNetTransition> PTNet::getEnabledTransitions(PTMarking marking) const
  {
    list<PTNetTransition> result;
    for(list<PTNetTransition>::const_iterator i = transitions.begin();
        i != transitions.end(); i++)
    {
      if (i->isEnabledBy(marking))
      {
        result.push_back(*i);
      }
    }
    return result;
  }  

  bool PTNet::isCoverableMarking(const PTMarking& marking)
  {
    if (ncov)
      return true;

    if (m_covGraph.m_nodeMapping.empty())
    {
      //first call at all or since last CovGraph::clear()
      m_covGraph.addNode(initial_marking);
      IDTYPE firstNode = m_covGraph.m_nodeMapping[initial_marking];
      m_covGraph.m_work.push_back(pair<IDTYPE,list<PTNetTransition> >(firstNode, getEnabledTransitions(initial_marking)));
    }
    
    for(map<PTMarking,IDTYPE>::iterator i = m_covGraph.m_nodeMapping.begin();
        i != m_covGraph.m_nodeMapping.end();
        i++)
    {
      if (isLower(marking, i->first))
      {
        return true;
      }
    }
    
    while (! m_covGraph.m_work.empty())
    {
      IDTYPE actualNode = m_covGraph.m_work.begin()->first;
      PTMarking actualMarking = m_covGraph.m_markingMapping[actualNode];
      
      if ((m_covGraph.m_work.begin()->second).empty())
      {
        m_covGraph.m_work.erase(m_covGraph.m_work.begin());
      }
      else
      {
        PTNetTransition actualTransition = *(m_covGraph.m_work.begin()->second).begin();
        (m_covGraph.m_work.begin()->second).erase((m_covGraph.m_work.begin()->second).begin());      
        
        PTMarking newMarking = actualTransition.fire(actualMarking);
        
        PTMarking smallerMarking;
        
        bool smaller = m_covGraph.getNodeWithSmallerMarking(actualMarking, newMarking, smallerMarking);
        
        if (smaller)
        {
          for(PTMarking::iterator i = smallerMarking.begin();
              i!= smallerMarking.end();
              i++)
          {
            unsigned smallerMarkingValue = (*i).second;
            PTMarking::iterator newMarkingIt = newMarking.find((*i).first);
            unsigned newMarkingValue = newMarkingIt->second;
            if(smallerMarkingValue < newMarkingValue)
              newMarking[i->first] = OMEGA;
          }
        }
        
        if (! m_covGraph.containsNode(newMarking))
        {     
          if (isStrictlyLower(marking, newMarking))
          {
            return true;
          }
          m_covGraph.addNode(newMarking);
          list<PTNetTransition> trans = getEnabledTransitions(newMarking);
          if (! trans.empty())
            m_covGraph.m_work.push_back(pair<IDTYPE,list<PTNetTransition> >(m_covGraph.m_nodeMapping[newMarking], trans));            
        }
        
        m_covGraph.addEdge(actualMarking, newMarking, actualTransition);
      
      
      }
    }
    return false;
  }
  

  PTNet::PTNet(const std::set<IDTYPE> p_places,
               const std::list<PTNetTransition> p_transitions,
               const PTMarking p_initial_marking)
    /*    : places(p_places),
      transitions(p_transitions),
      initial_marking(p_initial_marking)*/
  {
    places = set<IDTYPE>(p_places);
    transitions = list<PTNetTransition>(p_transitions);
    initial_marking = PTMarking(p_initial_marking);
    current_marking = PTMarking(p_initial_marking);
  }
  
    
  PTNet::PTNet(const PTNet& net)
    :places(net.places),
     transitions(net.transitions),
     initial_marking(net.initial_marking),
     current_marking(net.current_marking)
  {
  }

  PTNet::PTNet()
  {
  }
  

  bool PTNet::inCausalityRelation(IDTYPE element1, IDTYPE element2)
  {
    set<IDTYPE> visited; // set of visited elements in DFS
    
    if(1==places.count(element1)) // element1 is a place
    {
      return inCausalityRelationPlace(element1,element2,visited);
    }
    else // element1 is a transition
    {
      return inCausalityRelationTransition(element1,element2,visited);
    }
  }

  bool PTNet::inCausalityRelationPlace(IDTYPE element1,
                                       IDTYPE element2,
                                       set<IDTYPE>& visited)
  {
    if(1==visited.count(element1))
      return false;
    
    for(list<PTNetTransition>::iterator i = transitions.begin();
        i!=transitions.end();
        i++)
    {
      // element1 in preset of element2?
      if((*i).getID() == element2 && 1==(*i).getPreset().count(element1))
        return true;
      }

      // start recursive search for a path from element1 to element2
      visited.insert(element1);
      for(list<PTNetTransition>::iterator i = transitions.begin();
          i!=transitions.end();
          i++)
      {
        if(1==(*i).getPreset().count(element1))  // *i has element1 in preset
        {
          if(inCausalityRelationTransition((*i).getID(),element2,visited))
            return true;
        }
      }
    return false;
  }

  bool PTNet::inCausalityRelationTransition(IDTYPE element1,
                                            IDTYPE element2,
                                            set<IDTYPE>& visited)
  {
    if(1==visited.count(element1))
      return false;

    for(list<PTNetTransition>::iterator i = transitions.begin();
        i!=transitions.end();
        i++)

      // *i has element2 in postset
      if(element1 == (*i).getID() && 1==(*i).getPostset().count(element2))
        return true;

      bool found = false;
      visited.insert(element1);
      for(list<PTNetTransition>::iterator i = transitions.begin();
          !found && i!=transitions.end();
          i++)
      {
        if(element1==(*i).getID())
        {
          found = true;
          set<IDTYPE> postset = (*i).getPostset();
          for(set<IDTYPE>::iterator j = postset.begin();
              j!=postset.end();
              j++)
          {
            if(inCausalityRelationPlace(*j,element2,visited))
              return true;
          }
        }
      }
    return false;
  }
  
  std::set<IDTYPE> PTNet::getPlaces()
  {
    return places;
  }
  
  std::list<PTNetTransition> PTNet::getTransitions()
  {
    return transitions;
  }

  PTNetTransition& PTNet::getTransition(IDTYPE id)
  {
    for(list<PTNetTransition>::iterator i = transitions.begin();
        i!=transitions.end();
        i++)
      if(id == (*i).getID())
        return *i;

    // do not search for non-existent transitions:
    assert(false);
  }
  
  void PTNet::addPlace(IDTYPE p)
  {
    places.insert(p);
    m_covGraph.clear();
  }
  void PTNet::addTransition(PTNetTransition t)
  {
    transitions.push_back(t);
    m_covGraph.clear();  
  }

  ostream& operator<<(ostream& os, PTNet& ptn)
  {
    os << "Place/Transition net \n";
    for(list<PTNetTransition>::iterator i = ptn.transitions.begin();
        i!= ptn.transitions.end();
        i++)
      os << "   " << (*i) << "\n";
    return os;
  }

  
  bool isEqual(const PTMarking& m1, const PTMarking& m2)
  {

    for(PTMarking::const_iterator i = m1.begin();
        i!= m1.end();
        i++)
    {
      unsigned found_marking = (*i).second;
      PTMarking::const_iterator loop_marking_it = m2.find((*i).first);
      unsigned loop_marking = (*loop_marking_it).second;
      if(found_marking != loop_marking) // mismatch
        return false;
    }
    return true;
  }

  // Corrected function isLower: a marking m is lower than itself (Barbara, 16-05-03)
  bool isLower(const PTMarking& m1, const PTMarking& m2)
  {
    for(PTMarking::const_iterator i=m1.begin();
        i != m1.end();
        i++)
    {
      unsigned m1_marking = (*i).second;
      PTMarking::const_iterator m2_marking_it = m2.find((*i).first);
      unsigned m2_marking = (*m2_marking_it).second;
      if(m1_marking>m2_marking)
        return false;
    }
    return true;
  }
  
  bool isStrictlyLower(const PTMarking& m1, const PTMarking& m2)
  {
    bool different = false; // set to true if any difference is found
     
    for(PTMarking::const_iterator i=m1.begin();
        i != m1.end();
        i++)
      {
        unsigned m1_marking = (*i).second;
        PTMarking::const_iterator m2_marking_it = m2.find((*i).first);
        unsigned m2_marking = (*m2_marking_it).second;
        if(m1_marking>m2_marking)
          return false;
        if(m1_marking != m2_marking)
          different = true;
      }
    return different;
     
  }
   
void PTNet::deleteTransition(IDTYPE id)
{
	bool found = false;

	list<PTNetTransition>::iterator it;
	
	for(it = transitions.begin(); it!=transitions.end(); it++)
      if(id == (*it).getID())
      {
				found = true;
        break;
      }

  if(found)
		transitions.erase(it);					
}

///////////////////////////////////////////
/** set methods */

/** id */
void PTNetTransition::setID(IDTYPE id)
{
	 this->id = id;
}

/** add a place to postset */
void PTNetTransition::addToPostset(IDTYPE place, int weight)
{
   postset.insert(place);
   
   postweight[place] = weight;
}

/** add a place to preset */
void PTNetTransition::addToPreset(IDTYPE place, int weight)
{
   preset.insert(place);

   preweight[place] = weight;

}

/** set rule */
void PTNetTransition::setAssociatedRule(IDTYPE rule)
{
   associatedRule = rule;
}

/** fire transition */
bool PTNet::fireTransition(IDTYPE trans_id)
{
	list<PTNetTransition>::iterator itTrans;

  for(itTrans = transitions.begin(); itTrans!=transitions.end(); itTrans++)
     if(itTrans->getID()==trans_id)
			 break;

    // can be fired ?
	  bool can_be = true;   
	  set<IDTYPE> preset = itTrans->getPreset();
    for(set<IDTYPE>::iterator it = preset.begin(); it != preset.end(); it++)
    {
			if(current_marking[*it] < itTrans->getPreWeight(*it))
			{
				can_be=false;
				break;
			}
		}

    if(can_be)
    {
      for(set<IDTYPE>::iterator it = preset.begin(); it != preset.end(); it++)
				current_marking[*it] -= itTrans->getPreWeight(*it);
		
		
			// postset
			set<IDTYPE> postset = itTrans->getPostset();

			for(set<IDTYPE>::iterator it = postset.begin(); it != postset.end(); it++)
				current_marking[*it] += itTrans->getPostWeight(*it);

			return true;
		}
		else
			return false;
}

/** firable transitions */
list<IDTYPE> PTNet::firableTransitions(IDTYPE id_rule)
{
  list<list<PTNetTransition>::iterator> listit;

  list<IDTYPE> firable;

  for(list<PTNetTransition>::iterator it = transitions.begin(); it!=transitions.end(); it++)
     if(it->getAssociatedRule()==id_rule)
			 listit.push_back(it);

	for( list<list<PTNetTransition>::iterator>::iterator itTrans = listit.begin();
					itTrans!=listit.end(); itTrans++)
	{
    // can be fired ?
	  bool can_be = true;
	  set<IDTYPE> preset = (*itTrans)->getPreset();
    for(set<IDTYPE>::iterator it = preset.begin(); it != preset.end(); it++)
    {
			if(current_marking[*it] < (*itTrans)->getPreWeight(*it))
			{
				can_be=false;
				break;
			}
		}

    if(can_be)
			firable.push_back((*itTrans)->getID());
	}

	return firable;
}



/** get id of the transition that can be fired */
set<IDTYPE> PTNet::getPossibleIds(IDTYPE id_rule)
{
  list<list<PTNetTransition>::iterator> listit;

  set<IDTYPE> result;

  for(list<PTNetTransition>::iterator it = transitions.begin(); it!=transitions.end(); it++)
     if(it->getAssociatedRule()==id_rule)
			 listit.push_back(it);

  for( list<list<PTNetTransition>::iterator>::iterator itTrans = listit.begin();
					itTrans!=listit.end(); itTrans++)
	{
		// can be fired ?
		bool can_be = true;
		set<IDTYPE> preset = (*itTrans)->getPreset();
		for(set<IDTYPE>::iterator it = preset.begin(); it != preset.end(); it++)
		{
			if(current_marking[*it] < (*itTrans)->getPreWeight(*it))
			{
				can_be=false;
				break;
			}
		}

		if(can_be)
			result.insert((*itTrans)->getID());
	}	

  return result;
}

/** Restore the intial marking in PN */
void PTNet::setCurrentMarking(PTMarking ptm)
{
	current_marking = ptm;
}
